import uuid
from datetime import datetime
from typing import Any, Optional

import pytest
from pydantic import BaseModel

import pys
from pys.pydantic import ModelWithID


class Created(BaseModel):
    created: str = None

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        if self.created is None:
            self.created = datetime.utcnow().isoformat()


class BotBrief(ModelWithID, Created):
    name: str
    enabled: bool = True


class Bot(BotBrief):
    token: Optional[str] = None
    owner_id: Optional[str] = None
    notification_chat_id: Optional[str] = None
    pages: Optional[Any] = None


class User(ModelWithID):
    username: str
    first_name: str
    last_name: str


class BotOwner(User):
    pass


class ResponseContent(BaseModel):
    content: Optional[Any] = None


class BotResponse(ModelWithID, ResponseContent, Created):
    author: User


storages = [
    pys.file_storage('tests.storage'),
    pys.sqlite_storage('tests.db'),
]


@pytest.fixture(params=storages)
def storage(request):
    return request.param


def test_load(storage):
    owner = BotOwner(username='itisme', first_name='Test', last_name='Tester')
    assert owner.id is None
    owner.__my_id__()
    assert str(uuid.UUID(owner.id, version=4))
    storage.save(owner)

    bot_owner = storage.load(BotOwner, owner.id)
    assert bot_owner is not None
    assert bot_owner.id == owner.id
    assert bot_owner.username == owner.username
    assert bot_owner.first_name == owner.first_name
    assert bot_owner.last_name == owner.last_name
    storage.delete(bot_owner.__class__, bot_owner.id)


def test_save(storage):
    bot_owner2a = BotOwner(first_name='Second', last_name='', username='')
    storage.save(bot_owner2a)
    bot_owner2b = storage.load(BotOwner, bot_owner2a.id)
    assert bot_owner2b == bot_owner2a
    storage.delete(BotOwner, bot_owner2a.id)


def test_related_model(storage):
    bot_owner2 = BotOwner(first_name='Second', last_name='Owner', username='')
    bot_owner3 = BotOwner(first_name='Third', last_name='Owner', username='')
    storage.save(Bot(id='3', name='A bot'), bot_owner2)
    storage.save(Bot(id='3', name='B bot'), bot_owner3)

    bot_a = storage.load(Bot, '3', bot_owner2)
    assert bot_a.id == '3'
    assert bot_a.name == 'A bot'

    bot_b = storage.load(Bot, '3', bot_owner3)
    assert bot_b.id == '3'
    assert bot_b.name == 'B bot'

    storage.delete(bot_owner2.__class__, bot_owner2.id)
    storage.delete(bot_owner3.__class__, bot_owner3.id)


def test_related_models(storage):
    bot_owner = BotOwner(first_name='Second', last_name='Owner', username='')
    bot = Bot(id='1', name='A bot')
    storage.save(bot, bot_owner)
    response = BotResponse(author=bot_owner, content='123')
    storage.save(response, bot_owner, bot)

    bot_response = storage.load(BotResponse, response.id, bot_owner, bot)
    assert bot_response.id == response.id
    assert bot_response.content == response.content

    bot_response = storage.load(BotResponse, response.id, bot_owner, (Bot, bot.id))
    assert bot_response.id == response.id
    assert bot_response.content == response.content

    storage.delete(bot_owner.__class__, bot_owner.id)


def test_list(storage):
    bot_owner = BotOwner(first_name='Second', last_name='Owner', username='')
    storage.save(Bot(id='1', name='Bot #1'), bot_owner)
    storage.save(Bot(id='2', name='Bot #2'), bot_owner)
    storage.save(Bot(id='3', name='Bot #3'), bot_owner)

    bot_list = list(storage.list(Bot, bot_owner))
    assert len(bot_list) == 3
    bot_ids = ('1', '2', '3')
    for bot in bot_list:
        assert bot.id in bot_ids

    storage.delete(bot_owner.__class__, bot_owner.id)


def test_tear_down(storage):
    storage.destroy()



import os
import re
import sqlite3
from pathlib import Path
from typing import Type, Iterable, Optional, Any

from .base import BaseStorage, StoredModel, Related


class Storage(BaseStorage):
    con = None

    def _ensure_table_exist(self, table_name):
        for cnt in self.con.execute(
                "select count(name) from sqlite_master where type='table' and name=?;",
                (table_name,)
        ):
            if not cnt[0] == 1:
                self.con.execute(
                    f"""
                    create table {table_name} (
                        id varchar(255) not null,
                        data json,
                        related_id varchar(255),
                        related_name varchar(255),
                        unique (id, related_id, related_name)
                    );

                    """
                )

    @staticmethod
    def _get_table_name(cls):
        return re.sub(r'\W', '_', cls.__name__).lower()

    def __init__(self, path: Path):
        self.base_path = Path(path)
        self.con = sqlite3.connect(self.base_path)

    @staticmethod
    def _related(related_model: Related):
        (prev_cls, prev_id) = None, None
        if related_model:
            if isinstance(related_model, tuple):
                (prev_cls, prev_id) = related_model
            else:
                (prev_cls, prev_id) = related_model.__class__, related_model.__my_id__()
        return prev_cls, prev_id

    def load(self, model_class: Type[StoredModel], model_id: str, *related_model: Related) -> Optional[StoredModel]:
        last_related = related_model[-1] if related_model else None
        (rel_cls, rel_id) = self._related(last_related)

        table_name = self._get_table_name(model_class)
        self._ensure_table_exist(table_name)

        for row in self.con.execute(
            f"""
            select id, data, related_id, related_name
            from {table_name}
            where 
                id=? and {'related_id=? and related_name=?' if rel_id else 'related_id is null'}
            """,
            (model_id, rel_id, rel_cls.__name__) if last_related else (model_id,),
        ):
            return model_class.__factory__(row[1], row[0])
        else:
            return None

    def _save(self, model: StoredModel, prev=Related):
        table_name = self._get_table_name(model.__class__)
        self._ensure_table_exist(table_name)

        (prev_cls, prev_id) = self._related(prev)
        self.con.execute(
            f"""
            insert into {table_name} (id, data, related_id, related_name)
            values (?, ?, ?, ?) 
            on conflict do update set data=excluded.data;
            """,
            (model.__my_id__(),
             model.__json__(),
             prev_id,
             prev_cls.__name__ if prev else None,),
        )
        return model.__my_id__()

    def save(self, model: StoredModel, *related_model: Related) -> Any:
        prev_model = None
        last_id = None
        for m in related_model + (model,):
            last_id = self._save(m, prev=prev_model)
            prev_model = m
        return last_id

    def delete(self, model_class: Type[StoredModel], model_id: str, *related_model: Related) -> None:
        last_related = related_model[-1] if related_model else None
        (prev_cls, prev_id) = self._related(last_related)

        table_name = self._get_table_name(model_class)
        self._ensure_table_exist(table_name)

        self.con.execute(
            f"""
            delete from {table_name}
            where id=? and related_id=? and related_name=? 
            """,
            (model_id, prev_id, prev_cls.__name__ if prev_cls else None,)
        )

    def list(self, model_class: Type[StoredModel], *related_model: Related) -> Iterable[StoredModel]:
        last_related = related_model[-1] if related_model else None
        (prev_cls, prev_id) = self._related(last_related)

        table_name = self._get_table_name(model_class)
        self._ensure_table_exist(table_name)

        return [model_class.__factory__(row[1], row[0]) for row in self.con.execute(
            f"""
            select distinct id, data, related_id, related_name
            from {table_name}
            where {'related_id=? and related_name=?' if prev_cls else 'related_id is null'}   
            """,
            (prev_id, prev_cls.__name__) if prev_cls else (),
        )]

    def destroy(self) -> None:
        self.con.close()
        if self.base_path.exists():
            os.unlink(self.base_path)

    def __str__(self) -> str:
        return f'sqlite.Storage(base_path={self.base_path})'



from dataclasses import dataclass
from typing import Any

import msgspec
import pytest
from pydantic import BaseModel
from pys.pydantic import ModelWithID

import pys


@pytest.fixture
def storage():
    storage = pys.file_storage('storage.db')
    yield storage
    storage.destroy()


def dataclass_cases() -> list[tuple]:
    @pys.saveable
    @dataclass
    class AuthorDcNoId:
        name: str

    @pys.saveable
    @dataclass
    class BookDcNoId:
        title: str

    @pys.saveable
    @dataclass
    class AuthorDcWithId:
        name: str
        id: str = None

    @pys.saveable
    @dataclass
    class BookDcWithId:
        title: str
        id: str = None

    @pys.saveable(field_as_id='name')
    @dataclass
    class AuthorDcWithCustomId:
        name: str

    @pys.saveable(field_as_id='title')
    @dataclass
    class BookDcWithCustomId:
        title: str

    return [
        (AuthorDcNoId, BookDcNoId),
        (AuthorDcWithId, BookDcWithId),
        (AuthorDcWithCustomId, BookDcWithCustomId),
    ]


def msgspec_structs() -> list[tuple]:
    @pys.saveable
    class AuthorStructNoId(msgspec.Struct):
        name: str

    @pys.saveable
    class BookStructNoId(msgspec.Struct):
        title: str

    @pys.saveable
    class AuthorStructWithId(msgspec.Struct):
        name: str
        id: str = None

    @pys.saveable
    class BookStructWithId(msgspec.Struct):
        title: str
        id: str = None

    @pys.saveable(field_as_id='name')
    class AuthorStructWithCustomId(msgspec.Struct):
        name: str

    @pys.saveable(field_as_id='title')
    class BookStructWithCustomId(msgspec.Struct):
        title: str

    return [
        (AuthorStructNoId, BookStructNoId),
        (AuthorStructWithId, BookStructWithId),
        (AuthorStructWithCustomId, BookStructWithCustomId),
    ]


def pydantic_models() -> list[tuple]:
    @pys.saveable
    class AuthorPdNoId(BaseModel):
        name: str

    @pys.saveable
    class BookPdNoId(BaseModel):
        title: str

    @pys.saveable
    class AuthorPdWithId(BaseModel):
        name: str
        id: str = None

    @pys.saveable
    class BookPdWithId(BaseModel):
        title: str
        id: str = None

    @pys.saveable(field_as_id='name')
    class AuthorPdWithCustomId(BaseModel):
        name: str

    @pys.saveable(field_as_id='title')
    class BookPdWithCustomId(BaseModel):
        title: str

    return [
        (AuthorPdNoId, BookPdNoId),
        (AuthorPdWithId, BookPdWithId),
        (AuthorPdWithCustomId, BookPdWithCustomId),
    ]


def models_with_id() -> list[tuple]:
    class AuthorModelWithId(ModelWithID):
        name: str

    class BookModelWithId(ModelWithID):
        title: str

    return [
        (AuthorModelWithId, BookModelWithId),
    ]


def custom_cases() -> list[tuple]:
    @pys.saveable(field_as_id='name')
    class CustomAuthor(pys.Persistent):
        name: str

        def __init__(self, name):
            self.name = name

        @classmethod
        def __factory__(cls, raw_content: str, model_id: Any) -> 'CustomAuthor':
            return CustomAuthor(raw_content)

        def __json__(self) -> str:
            return self.name

        def __eq__(self, o: object) -> bool:
            return self.name == o.name

    @pys.saveable(field_as_id='title')
    class CustomBook(pys.Persistent):
        title: str

        def __init__(self, title):
            self.title = title

        @classmethod
        def __factory__(cls, raw_content: str, model_id: Any) -> 'CustomBook':
            return CustomBook(raw_content)

        def __json__(self) -> str:
            return self.title

        def __eq__(self, o: object) -> bool:
            return self.title == o.title

    return [
        (CustomAuthor, CustomBook),
    ]


@pytest.mark.parametrize(
    argnames=('cls_author', 'cls_book'),
    argvalues=dataclass_cases() + msgspec_structs() + pydantic_models() + models_with_id() + custom_cases()
)
def test_sample(storage, cls_author, cls_book):
    leo = cls_author(name='Leo Tolstoy')
    storage.save(leo)
    assert leo.__my_id__()

    leo2 = storage.load(cls_author, leo.__my_id__())
    assert leo2.name == leo.name

    war_and_peace = cls_book(title='War and peace')
    storage.save(war_and_peace, leo)
    gpt = cls_author(name='Chat GPT')
    storage.save(gpt)
    gpt_war_and_peace = storage.load(cls_book, war_and_peace.__my_id__(), gpt)
    assert gpt_war_and_peace is None

    storage.save(war_and_peace, gpt)
    gpt_war_and_peace = storage.load(cls_book, war_and_peace.__my_id__(), gpt)
    assert gpt_war_and_peace is not None

    for_kids = cls_book(title='For Kids')
    storage.save(for_kids, leo)
    leo_books = list(storage.list(cls_book, leo))
    assert len(leo_books) == 2
    assert war_and_peace in leo_books
    assert for_kids in leo_books



from pydantic import BaseModel

from . import saveable


@saveable(field_as_id='id')
class ModelWithID(BaseModel):
    """
    Base class for models with `id` field prefilled with random UUID if not initialized.
    """
    id: str = None



import abc
import functools
import uuid
from contextlib import contextmanager
from pathlib import Path
from typing import Union, Callable, Any

from . import file, sqlite


def _random_uuid(_) -> str:
    return str(uuid.uuid4())


def _is_dataclass(cls):
    import dataclasses
    return dataclasses.is_dataclass(cls)


def _is_pydantic(cls):
    try:
        from pydantic import BaseModel
        return issubclass(cls, BaseModel)
    except:
        return False


def _is_msgspec_struct(cls):
    try:
        import msgspec
        return issubclass(cls, msgspec.Struct)
    except:
        return False


class Persistent(abc.ABC):
    """
    Base class for any object ready to be persisted
    """
    @classmethod
    def __factory__(cls, raw_content: str, model_id: Any) -> 'Persistent':
        """
        Create object of class from `raw_content` with ID from `model_id`
        :param raw_content: Raw content
        :param model_id: Model ID
        :return: Instance of class
        """
        raise NotImplementedError

    def __json__(self) -> str:
        """
        Get JSON representation of the object
        :return: JSON representation
        """
        raise NotImplementedError


def saveable(base_cls=None, *,
             field_as_id: str = 'id',
             default_id: Callable[[Any], str] = _random_uuid):
    """
    Decorate the given `cls` with `__my_id__()` and `__json__()` methods
    required for persistence.
    :param base_cls: Class to decorate.
    :param field_as_id: existing class field to be used as object ID.
    :param default_id: Default ID value function (id(self) by default).
    :return: Decorated class
    """
    if not base_cls:
        def wrapper(decor_cls):
            return saveable(decor_cls, field_as_id=field_as_id)

        return wrapper

    @functools.wraps(base_cls, updated=())
    class _BasePersistence(base_cls):
        @classmethod
        def __factory__(cls, raw_content: str, model_id: Any) -> base_cls:
            if hasattr(base_cls, '__factory__'):
                return base_cls.__factory__(raw_content, model_id)
            else:
                raise NotImplementedError(
                    f'The class {base_cls} is not msgspec.Struct, @dataclass nor Pydantic Model '
                    f'and does not have __factory__() method. Please implement __factory__() method by yourself.')

        @staticmethod
        def _valid_id(_id: Any) -> Any:
            if not _id:
                raise ValueError('ID shall not be empty')
            return _id

        def _original_id(self) -> Union[Any, None]:
            if hasattr(base_cls, '__my_id__') and callable(cls_my_id := getattr(base_cls, '__my_id__')):
                return cls_my_id(self)
            return None

        def __my_id__(self) -> Any:
            """
            Get object ID to be used for persisting. If `field_as_id` is specified
            then use this field as ID.
            :return: Object's ID
            """
            if original_id := self._original_id():
                return original_id
            _id = getattr(self, field_as_id, None)
            if not _id:
                _id = default_id(self)
                setattr(self, field_as_id, _id)
            return self._valid_id(_id)

        def __json__(self) -> str:
            """
            Get JSON representation of the object
            :return: JSON representation
            """
            if hasattr(base_cls, '__json__'):
                return base_cls.__json__(self)
            else:
                raise NotImplementedError(
                    f'The class {base_cls} is not msgspec.Struct, @dataclass nor Pydantic Model '
                    f'and does not have __json__() method. Please implement __json__() method by yourself.')

    @functools.wraps(base_cls, updated=())
    class _HasMyIdMethod(_BasePersistence):
        def __my_id__(self) -> str:
            return self._original_id()

    @functools.wraps(base_cls, updated=())
    class _NoIdField(_BasePersistence):
        @classmethod
        def _parent_factory(cls) -> Callable[[str, Any], _BasePersistence]:
            raise NotImplementedError

        def set_saved_id(self, model_id: Any):
            self.__my_saved_id__ = model_id

        @classmethod
        def __factory__(cls, raw_content: str, model_id: str) -> '_NoIdField':
            model = cls._parent_factory()(raw_content, model_id)
            model.set_saved_id(model_id)
            return model

        def __my_id__(self) -> str:
            if original_id := self._original_id():
                return original_id
            if not self.__my_saved_id__:
                self.set_saved_id(default_id(self))
            return self._valid_id(self.__my_saved_id__)

    has_id = hasattr(base_cls, field_as_id)
    has_my_id = hasattr(base_cls, '__my_id__') and callable(getattr(base_cls, '__my_id__'))
    parent = _HasMyIdMethod if has_my_id else _BasePersistence

    if _is_pydantic(base_cls):
        has_id = field_as_id in base_cls.__annotations__

        @functools.wraps(base_cls, updated=())
        class _Pydantic(parent):
            @classmethod
            def __factory__(cls, raw_content: str, model_id: Any) -> '_Pydantic':
                return cls.model_validate_json(raw_content)

            def __json__(self) -> str:
                return self.model_dump_json()

        @functools.wraps(base_cls, updated=())
        class _PydanticNoId(_Pydantic, _NoIdField):
            # __slots__ = (MY_SAVED_ID,)
            __my_saved_id__ = None

            @classmethod
            def _parent_factory(cls) -> Callable[[str, Any], _Pydantic]:
                return _Pydantic.factory

        return _Pydantic if has_id or has_my_id else _PydanticNoId

    @functools.wraps(base_cls, updated=())
    class _MsgspecStruct(parent):
        @classmethod
        def __factory__(cls, raw_content: str, model_id: Any) -> '_MsgspecStruct':
            import msgspec
            content = msgspec.json.decode(raw_content)
            return cls(**content)

        def _prepare(self) -> dict:
            return self

        def __json__(self) -> str:
            import msgspec
            return msgspec.json.encode(self._prepare()).decode(encoding='UTF-8')

    @functools.wraps(base_cls, updated=())
    class _MsgspecStructNoIdField(_MsgspecStruct, _NoIdField):
        import msgspec
        __my_saved_id__: Union[str, None, msgspec.UnsetType] = msgspec.UNSET

        @contextmanager
        def without_saved_id(self):
            import msgspec
            copy = self.__my_saved_id__
            try:
                self.__my_saved_id__ = msgspec.UNSET
                yield self
            finally:
                self.__my_saved_id__ = copy

        def __eq__(self, other: '_MsgspecStructNoIdField'):
            if isinstance(other, base_cls):
                with self.without_saved_id(), other.without_saved_id() as _other:
                    return super().__eq__(_other)
            else:
                return super().__eq__(other)

        @classmethod
        def _parent_factory(cls) -> Callable[[str, Any], _MsgspecStruct]:
            return _MsgspecStruct.factory

        def __json__(self) -> str:
            with self.without_saved_id():
                return super().__json__()

    if _is_msgspec_struct(base_cls):
        return _MsgspecStruct if has_id or has_my_id else _MsgspecStructNoIdField

    if _is_dataclass(base_cls):
        has_id = field_as_id in base_cls.__annotations__
        from dataclasses import asdict

        @functools.wraps(base_cls, updated=())
        class _Dataclass(_MsgspecStruct):
            def _prepare(self) -> dict:
                # noinspection PyDataclass
                return asdict(self)

        @functools.wraps(base_cls, updated=())
        class _DataclassNoId(_Dataclass, _NoIdField):
            # __slots__ = ('__my_saved_id__',)
            __my_saved_id__ = None

            @classmethod
            def _parent_factory(cls) -> Callable[[str, Any], _Dataclass]:
                return _Dataclass.factory

        return _Dataclass if has_id or has_my_id else _DataclassNoId

    return _HasMyIdMethod if has_my_id else _BasePersistence


def file_storage(base_path: Union[str, Path]):
    return file.Storage(base_path)


def sqlite_storage(base_path: Union[str, Path]):
    return sqlite.Storage(base_path)


storage = sqlite_storage

__all__ = ('saveable', 'storage', 'file_storage', 'sqlite_storage', 'Persistent')



import re
from dataclasses import dataclass

import msgspec
from pydantic import BaseModel

import pys


def test_dataclass():
    @dataclass
    class Url:
        url: str

        @staticmethod
        def to_id(url: str) -> str:
            NON_ALNUM = re.compile(r'([^a-zA-Z0-9_]+)')
            return NON_ALNUM.sub('_', url)

        def __my_id__(self):
            return Url.to_id(self.url)

    @pys.saveable
    @dataclass
    class Link(Url):
        a: int = 0

    link = Link('https://example.com/')
    assert link.__my_id__() == Url.to_id(link.url)


def test_msgspec():
    class Url(msgspec.Struct):
        url: str

        @staticmethod
        def to_id(url: str) -> str:
            NON_ALNUM = re.compile(r'([^a-zA-Z0-9_]+)')
            return NON_ALNUM.sub('_', url)

        def __my_id__(self):
            return Url.to_id(self.url)

    @pys.saveable
    class Link(Url):
        a: int = 0

    link = Link('https://example.com/')
    assert link.__my_id__() == Url.to_id(link.url)


def test_pydantic():
    class Url(BaseModel):
        url: str

        @staticmethod
        def to_id(url: str) -> str:
            NON_ALNUM = re.compile(r'([^a-zA-Z0-9_]+)')
            return NON_ALNUM.sub('_', url)

        def __my_id__(self):
            return Url.to_id(self.url)

    @pys.saveable
    class Link(Url):
        a: int = 0

    link = Link(url='https://example.com/')
    assert link.__my_id__() == Url.to_id(link.url)



import pathlib

from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / 'README.md').read_text()

PYTEST_VERSION = '8.22'
PYDANTIC_VERSION = '2.0'
FILELOCK_VERSION = '3.15.4'
MSGSPEC_VERSION = '0.18.6'

setup(name='pysdato',
      version='0.0.12',
      python_requires='>=3.9',
      description='Simple JSON file storage for Python dataclasses, msgspec structs and pydantic models, thread and '
                  'multiprocess safe',
      long_description=README,
      long_description_content_type="text/markdown",
      url="https://github.com/stasdavydov/pys",
      author="Stas Davydov",
      author_email="davidovsv@yandex.ru",
      license="MIT",
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Intended Audience :: Developers",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.9",
          "Programming Language :: Python :: 3.10",
          "Programming Language :: Python :: 3.11",
          "Programming Language :: Python :: 3.12",
          "Topic :: Database :: Database Engines/Servers",
          "Topic :: Software Development :: Libraries",
          "Topic :: Software Development :: Libraries :: Python Modules",
      ],
      packages=['pys'],
      include_package_data=False,
      install_requires=[
          f'filelock >= {FILELOCK_VERSION}',
          f'msgspec >= {MSGSPEC_VERSION}',
      ],
      extras_require={
          'test': [
              f'pytest >= {PYTEST_VERSION}',
              f'msgspec >= {MSGSPEC_VERSION}',
              f'pydantic >= {PYDANTIC_VERSION}',
          ],
          'dataclass': [
              f'msgspec >= {MSGSPEC_VERSION}',
          ],
          'msgspec': [
              f'msgspec >= {MSGSPEC_VERSION}',
          ],
          'pydantic': [
              f'pydantic >= {PYDANTIC_VERSION}',
          ],
      }
      )



import abc
from typing import TypeVar, Union, Tuple, Type, Optional, Any, Iterable

StoredModel = TypeVar('StoredModel')
RelatedModel = TypeVar('RelatedModel')
Related = Union[RelatedModel, Tuple[RelatedModel, str]]


class BaseStorage(abc.ABC):
    """
    Abstract base storage
    """
    def load(self, model_class: Type[StoredModel], model_id: str,
             *related_model: Related) \
            -> Optional[StoredModel]:
        """
        Load model.

        :param model_class: Class of the model.
        :param model_id: Model ID.
        :param related_model: Related model(s) -- model that the loaded model is belong to.
        :return: Loaded model or None in case if model is not found.
        """
        raise NotImplementedError

    def save(self, model: StoredModel,
             *related_model: Related) -> Any:
        """
        Save model.

        :param model: A model.
        :param related_model: Related model(s) -- model that the stored model is belong to.
        :return Returns saved model id
        """
        raise NotImplementedError

    def delete(self, model_class: Type[StoredModel], model_id: str,
               *related_model: Related) -> None:
        """
        Delete model.

        :param model_class: Model class.
        :param model_id: Model ID.
        :param related_model: Related model(s) -- model that the deleted model is belong to.
        """
        raise NotImplementedError

    def list(self, model_class: Type[StoredModel],
             *related_model: Related) -> Iterable[StoredModel]:
        """
        List models.
        :param model_class: Model class
        :param related_model: Related model(s) -- model that the listed models are belong to.
        :return: List of found models.
        """
        raise NotImplementedError

    def destroy(self) -> None:
        """
        Destroy storage
        """
        raise NotImplementedError



import os
import shutil
from pathlib import Path
from typing import Type, Optional, Tuple, Iterable, Any, Union

from filelock import FileLock

from .base import BaseStorage, StoredModel, RelatedModel, Related


class Storage(BaseStorage):
    """
    File based storage implementation. Thread and interprocess safe.
    """
    base_path: Path

    def __init__(self, base_path: Union[str, Path]) -> None:
        """
        Base path for the storage files
        :param base_path: base path.
        """
        self.base_path = base_path if isinstance(base_path, Path) else Path(base_path)

    @staticmethod
    def _get_model_path(model_class: Type[StoredModel], model_id: str,
                        *related_model: Union[RelatedModel, Tuple[Type[RelatedModel], str]]) -> Path:
        path = Path()
        if not related_model:
            path /= ''
        else:
            for model in related_model:
                if isinstance(model, tuple):
                    path /= Storage._get_model_path(*model)
                else:
                    path /= Storage._get_model_path(model.__class__, model.__my_id__())

        return path / model_class.__name__ / model_id

    def _prepare_file(self, model_class: Type[StoredModel], model_id: str,
                      *related_model: Related):
        path = self.base_path / Storage._get_model_path(model_class, model_id, *related_model).with_suffix('.json')
        lock = path.with_suffix('.lock')
        path.parent.mkdir(parents=True, exist_ok=True)
        return path, FileLock(lock)

    def save(self, model: StoredModel,
             *related_model: Related) -> Any:
        model_id = model.__my_id__()
        path, lock = self._prepare_file(model.__class__, model_id, *related_model)
        with lock:
            path.write_text(model.__json__(), encoding='utf-8')
            return model_id

    def load(self, model_class: Type[StoredModel], model_id: str,
             *related_model: Related) -> Optional[StoredModel]:
        path, lock = self._prepare_file(model_class, model_id, *related_model)
        with lock:
            if not path.exists():
                return None
            return model_class.__factory__(path.read_text(encoding='utf-8'), model_id)

    def delete(self, model_class: Type[StoredModel], model_id: str,
               *related_model: Related) -> None:
        path, lock = self._prepare_file(model_class, model_id, *related_model)
        with lock:
            path.unlink(missing_ok=True)
            sub_path = path.with_suffix('')
            if sub_path.exists():
                shutil.rmtree(sub_path)

    _JSON_EXT_END = -5

    def list(self, model_class: Type[StoredModel],
             *related_model: Related) -> Iterable[StoredModel]:
        path, lock = self._prepare_file(model_class, '__list__', *related_model)
        with lock:
            for p in os.listdir(path.parent):
                if p.endswith('.json'):
                    yield self.load(model_class, p[:Storage._JSON_EXT_END], *related_model)

    def __str__(self) -> str:
        return f'file.Storage(base_path={self.base_path})'

    def destroy(self) -> None:
        shutil.rmtree(self.base_path)
